#ifndef ANGLE_H
#define ANGLE_H
#include "object.h"
#include "pv.h"
#include "ppoly.h"

namespace acp {

class RefCount {
  int refCount;
public:
  RefCount () : refCount(0) {}
  virtual ~RefCount () {}
  void incRef () { refCount++; }
  void decRef () { if (--refCount == 0) delete this; }
};

// An angle.  Equivalent to point on unit circle.
class Angle {
  class RandomBranch : public Object<PV2> {
    void calculate () { PV2 b = get(); b.x = -1/b.y; set(b); }
  public:
    RandomBranch () {
      Parameter s1 = Parameter::constant(randomNumber(0.75, 1));
      Parameter s0 = -1/s1;
      set(PV2(s0, s1));
    }
  } static branch;
public:
  static PV2 getBranch () { return branch.get(); }

  enum Type { ZERO, INPUT, EXPLICIT, IMPLICIT, MEAN };
  virtual Type getType () = 0;

  virtual ~Angle () {}

  static Parameter u2sf (PV2 u, bool &flag) {
    flag = (u.x > 0);
    Parameter s = flag ? u.y / (1 + u.x) : -u.y / (1 - u.x);
    PV2 b = getBranch();
    if (s < b.x || s > b.y) {
      s = -1/s;
      flag = !flag;
    }
    return s;
  }

  // One of these virtual functions has to be overridden!!

  // (cos a, sin a)
  virtual PV2 getU () {
    bool flag;
    Parameter s = getS(flag);
    Parameter d = 1 + s*s;
    return flag ? PV2((1 - s*s)/d, 2*s/d) : PV2((s*s - 1)/d, -2*s/d);
  }

  // Rational parameterization in range branch.get().x to branch.get().y
  // flag is true if branch contains angle 0.
  virtual Parameter getS (bool &flag) { return u2sf(getU(), flag); }

  Parameter getS () { bool flag; return getS(flag); }
  bool getFlag () { bool flag; getS(flag); return flag; }

  virtual bool identical (Angle *that) { return this == that; }

  // -1 if 0 <= a < b < 2 PI
  // 0 if equal or identical
  Predicate2(Compare, Angle*, a, Angle*, b);

  // 1 if going from a to c through b is counterclockwise
  // 0 if a same as b or b same as c
  Predicate3(Circulation, Angle*, a, Angle*, b, Angle*, c);
};

class ZeroAngle : public Angle {
public:
  Type getType () { return ZERO; }
  PV2 getU () { return PV2::constant(1, 0); }
  Parameter getS (bool &flag) { flag = true; return Parameter::constant(0); }
  bool identical (Angle *that) { return that->getType() == ZERO; }
} extern angle0;

// Angle defined by parameter s and flag.
class InputAngle : public Angle, public Object<Parameter> {
public:
  virtual Type getType () { return INPUT; }
  // Create from angle in radians (approximately).
  InputAngle (double angle) {
    assert(!Parameter::isEnabled());
    set(u2sf(PV2::constant(cos(angle), sin(angle)), flag));
  }
  virtual void calculate () {}
  virtual Parameter getS (bool &flag) { flag = this->flag; return get(); }

private:
  bool flag;
};

class RootAngle;

class AnglePoly : public RefCount {
public:
  AnglePoly () {}
  virtual PPoly2 getPoly () = 0; // PPoly2 in x (cos theta) and y (sin theta)
  virtual bool identical (AnglePoly *that) { return this == that; }
  std::vector<Angle*> getRoots ();
  enum Type { EF, EF2, VF, EE, EFF, FFF, NNN, FFFF, U };
  virtual Type getType () = 0;
  Parameter polish (Parameter s, bool flag);

private:
  std::vector<Parameter> getRoots (PPoly2 &poly, bool flag);
  std::vector<Angle*> getExplicitAngles (PPoly2 &poly);
  PPoly1 univariate (PPoly2 &poly, bool flag);
  void term (const Parameter &u, int c, int s, int d, bool flag,
             std::vector<Parameter> &a);
};

class RootAngle : public Angle {
  friend class AnglePoly;
  AnglePoly *poly;
  int index;
public:
  RootAngle (AnglePoly *poly) : poly(poly), index(-1) { poly->incRef(); }
  ~RootAngle () { poly->decRef(); } 

  AnglePoly *getPoly () { return poly; }

  bool identical (Angle *thatAngle) {
    if (Angle::identical(thatAngle)) return true;
    RootAngle *that = dynamic_cast<RootAngle*>(thatAngle);
    if (that == 0) return false;
    if (this->index != that->index) return false;
    return this->poly->identical(that->poly);
  }

  Predicate2(Sign, AnglePoly*, poly, Angle*, root);
};

class ImplicitAngle : public RootAngle, public Object<Parameter> {
  friend class AnglePoly;
  void calculate () { set(getPoly()->polish(get(), flag)); }
  ImplicitAngle (Parameter s, bool flag, AnglePoly *poly)
    : flag(flag), RootAngle(poly) {
    set(s);
  }
public:
  virtual Type getType () { return IMPLICIT; }
  virtual Parameter getS (bool &flag) { flag = this->flag; return get(); }
private:
  bool flag;
};

class ExplicitAngle : public RootAngle, public Object<PV2> {
  friend class AnglePoly;
  void calculate ();
  // For a x + b y + c = 0, flag true/false means p +/- s (-b, a).
  ExplicitAngle (const PV2 &u, bool flag, AnglePoly *poly)
    : flag(flag), RootAngle(poly) {
    set(u);
  }
public:
  virtual Type getType () { return EXPLICIT; }
  virtual PV2 getU () { return get(); }
private:
  bool flag;
};

class MeanAngle : public Angle, public Object<PV2> {
  Angle *u, *v;
  double t;
  void calculate () {
    PV2 m = u->getU() * t + v->getU() * (1-t);
    set(m / m.dot(m).sqrt());
  }
public:
  Type getType () { return MEAN; }
  MeanAngle (Angle *u, Angle *v, double t) : u(u), v(v), t(t) {}
  PV2 getU () { return get(); }
};

}

#endif
