/*
  ACP (Adaptive Controlled Precision) Library
  for robust computational geometry

  Copyright (c) 2013-07-15
  Victor Joseph Milenkovic
  University of Miami
  vjm@miami.edu
  Elisha Sacks
  Purdue University
  eps@cs.purdue.edu

  This file contains code described in

Robust Complete Path Planning in the Plane
Victor Milenkovic, Elisha Sacks, and Steven Trac
Proceedings of the Workshop on the Algorithmic Foundations of Robotics (WAFR)
pages 37-52, 2012

   This code is under development.
   It is free for use for academic and research purposes.
   Permission from the authors is required for commercial use.
*/

#ifndef OBJECT_H
#define OBJECT_H

#include "pv.h"
#include <vector>

namespace acp {

class BaseObject {
  friend class Predicate;
  template<class P> friend class Object;
  friend class Poly;

  static std::vector<BaseObject*> precisionIncreased;

  virtual void decreasePrecision () = 0;

public:
  static void decreaseAll () {
    static int count;
    count++;
    for (int i = 0; i < precisionIncreased.size(); i++)
      precisionIncreased[i]->decreasePrecision();
    precisionIncreased.clear();
  }
};

template<class P>
class Object : BaseObject {
  friend class Poly;

  P p;
  
  int size () { return sizeof(P) / sizeof(Parameter); }

  Parameter &get (int i) { return ((Parameter *) &p)[i]; }

  bool increased () { return get(0).increased(); }

  void increasePrecision () {
    // std::cout << "increasing precision of " << this << std::endl;
    for (int i = 0; i < size(); i++)
      get(i).increasePrecision();
  }

  void decreasePrecision () {
    // std::cout << "decreasing precision of " << this << std::endl;
    for (int i = 0; i < size(); i++)
      get(i).decreasePrecision();
  }

public:
  bool uninitialized () { return get(0).uninitialized(); }

  // Object (int n) { assert(sizeof(P) == n * sizeof(Parameter)); }
  const P &get () {
    if (uninitialized()) {
      calculate();
      if (Parameter::highPrecision > 53u)
        precisionIncreased.push_back(this);
    }
    else if (!increased()) {
      increasePrecision();
      precisionIncreased.push_back(this);
      calculate();
    }
    return p;
  }

  void set (const P &p) {
    bool increasedBefore = increased();
    this->p = p;
    bool increasedAfter = increased();
    if (increasedBefore < increasedAfter)
      precisionIncreased.push_back(this);
  }

  virtual void calculate () = 0;

  void safe_calculate () {
    if (!Parameter::handleSignException) {
      calculate();
      return;
    }
    bool failed = false;
    while (true)
      try {
        calculate();
        if (failed) {
          decreasePrecision();
          decreaseAll();
        }
        return;
      } catch (SignException se) {
        if (Parameter::highPrecision < 212)
          Parameter::highPrecision = 212;
        failed = true;
      }
  }

  void really_safe_calculate () {
    bool failed = false;
    while (true)
      try {
        calculate();
        if (failed) {
          decreasePrecision();
          decreaseAll();
        }
        return;
      } catch (SignException se) {
        if (Parameter::highPrecision < 212)
          Parameter::highPrecision = 212;
        failed = true;
      }
  }
};

class Predicate {
  // Calculate the sign from the objects.
  virtual int sign () = 0;
  
public:
  operator int () {
    if (!Parameter::handleSignException)
      return sign();
    bool failed = false;
    while (true)
      try {
        int s = sign();
        if (failed)
          BaseObject::decreaseAll();
        return s;
      } catch (SignException se) {
        if (Parameter::highPrecision < 212)
          Parameter::highPrecision = 212;
        failed = true;
      }
  }

  int unsafe () { return sign(); }
};

}  

// Macro to create one-argument predicate.
#define Predicate1(P, t1, v1)                                    \
  class P : public acp::Predicate {                              \
    t1 v1;                                                       \
  int sign ();                                                   \
  public:                                                        \
  P (t1 v1) : v1(v1) {}                                          \
  };

// Macro to create two-argument predicate.
#define Predicate2(P, t1, v1, t2, v2)                            \
  class P : public acp::Predicate {                              \
    t1 v1; t2 v2;                                                \
  int sign ();                                                   \
  public:                                                        \
  P (t1 v1, t2 v2) : v1(v1), v2(v2) {}                           \
  };

// Macro to create three-argument predicate.
#define Predicate3(P, t1, v1, t2, v2, t3, v3)                       \
  class P : public acp::Predicate {                                 \
    t1 v1; t2 v2; t3 v3;                                            \
  int sign ();                                                      \
  public:                                                           \
  P (t1 v1, t2 v2, t3 v3) : v1(v1), v2(v2), v3(v3) {}               \
  };

// Macro to create four-argument predicate.
#define Predicate4(P, t1, v1, t2, v2, t3, v3, t4, v4)                 \
  class P : public acp::Predicate {                                   \
    t1 v1; t2 v2; t3 v3; t4 v4;                                       \
    int sign ();                                                      \
  public:                                                             \
  P (t1 v1, t2 v2, t3 v3, t4 v4) : v1(v1), v2(v2), v3(v3), v4(v4) {}  \
  };

// Macro to create five-argument predicate.
#define Predicate5(P, t1, v1, t2, v2, t3, v3, t4, v4, t5, v5)           \
  class P : public acp::Predicate {                                     \
    t1 v1; t2 v2; t3 v3; t4 v4; t5 v5;                                  \
    int sign ();                                                        \
  public:                                                               \
  P (t1 v1, t2 v2, t3 v3, t4 v4, t5 v5)                                 \
    : v1(v1), v2(v2), v3(v3), v4(v4), v5(v5) {}                         \
  };

#endif
