#include "ppoly.h"
using namespace std;
using namespace acp;

namespace acp {

Parameter PPoly1::polish (const Parameter &x) const
{
  Parameter xubP = x.ubP();
  int sub = value(xubP).sign();
  return shrinkBracket(x, sub);
}

PPoly1::PPoly1 (int d, Parameter *a) : d(d)
{
  this->a = new Parameter [d + 1];
  if (a)
    for (int i = 0; i <= d; ++i)
      this->a[i] = a[i];
}

PPoly1::PPoly1 (const PPoly1 &f) : d (f.d)
{
  a = new Parameter [d + 1];
  for (int i = 0; i <= d; ++i)
    a[i] = f.a[i];
}

PPoly1 & PPoly1::operator= (const PPoly1 &f)
{
  delete [] a;
  d = f.d;
  a = new Parameter [d+1];
  for (int i = 0; i <= d; ++i)
    a[i] = f.a[i];
  return *this;
}

Parameter PPoly1::value (const Parameter &x) const
{
  Parameter y = a[d];
  for (int i = d - 1; i >= 0; --i)
    y = x*y + a[i];
  return y;
}

Parameter PPoly1::der (const Parameter &x) const
{
  Parameter y = d*a[d];
  for (int i = d - 1; i > 0; --i)
    y = x*y + i*a[i];
  return y;
}

PPoly1 PPoly1::der () const
{
  PPoly1 g(d - 1);
  for (int i = 1; i <= d; ++i)
    g.a[i-1] = i*a[i];
  return g;
}

vector<Parameter> PPoly1::roots (Parameter lb, Parameter ub) const {
  if (d == 2)
    return roots2(lb, ub);
  if (d < 2)
    return roots1(lb, ub);
  PPoly1 df = der();
  vector<Parameter> dres = df.roots(lb, ub);
  int dn = dres.size();
  vector<Parameter> res;
  Parameter lbi = lb;
  int slbi = value(lbi).sign();
  for (int i = 0; i <= dn; ++i) {
    Parameter ubi = i < dn ? dres[i] : ub;
    int subi = value(ubi).sign();
    if (slbi == -subi) {
      Parameter x = lbi.interval(ubi);
      res.push_back(shrinkBracket(x, subi));
    }
    lbi = ubi;
    slbi = subi;
  }
  return res;
}

vector<Parameter> PPoly1::roots1 (Parameter lb, Parameter ub) const {
  vector<Parameter> res;
  if (d == 0)
    return res;
  if (d == 1) {
    Parameter r = -a[0]/a[1];
    if (lb < r && r < ub)
      res.push_back(r);
    return res;
  }
  assert(0);
  return res;
}

vector<Parameter> PPoly1::roots2 (Parameter lb, Parameter ub) const {
  vector<Parameter> res;
  const Parameter &A = a[2], &B = a[1], &C = a[0];
  Parameter D2 = B*B - 4.0*A*C;
  if (D2.sign() == -1)
    return res;
  Parameter D = D2.sqrt();
  Parameter r[2];

  // (-B - D)/2A < (-B + D)/2A if A>0
  bool Apos = (A.sign() == 1);
  if (B.sign() == -1) {
    Parameter DmB = D-B;
    r[!Apos] = 2.0*C/DmB;
    r[Apos] = 0.5*DmB/A;
  }
  else {
    Parameter DpB = D+B;
    r[!Apos] = -0.5*DpB/A;
    r[Apos] = -2.0*C/DpB;
  }

  if (lb < r[0] && r[0] < ub)
    res.push_back(r[0]);
  if (lb < r[1] && r[1] < ub)
    res.push_back(r[1]);

  return res;
}

Parameter PPoly1::shrinkBracket (Parameter x, int sub) const {
  bool bflag = false;
  while (true) {
    bool flag = true;
    Parameter xm = x.midP(), fm = value(xm), df = der(x);
    if (df.sign(false) == 0)
      flag = false;
    else {
      Parameter nx = xm - fm/df;
      if (nx.subset(x))
	x = nx;
      else
	flag = false;
    }
    if (!flag) {
      int sm = fm.sign(false);
      if (sm == sub) {
	Parameter nx = x.interval(xm);
	if (nx.intervalWidth() == x.intervalWidth())
	  break;
	x = nx;
      }
      else if (sm == - sub) {
	Parameter nx = xm.interval(x);
	if (nx.intervalWidth() == x.intervalWidth())
	  break;
	x = nx;
      }
      else if (!bflag) {
	bflag = true;
	x = shrinkBracketB(x);
      }
      else
	break;
    }
  }
  return x;
}

Parameter PPoly1::shrinkBracketB (const Parameter &x) const {
  Parameter xlb = x.lbP(), xm = x.midP(), xub = x.ubP();
  while ((xlb - xm).sign(false) < 0) {
    Parameter nx = xlb.interval(xm).midP();
    if ((nx - xlb).sign(false) == 0 || (nx - xm).sign(false) == 0)
      break;
    int sm = value(nx).sign(false);
    if (sm == 0)
      xm = nx;
    else
      xlb = nx;
  }
  xm = x.midP();
  while ((xm - xub).sign(false) < 0) {
    Parameter nx = xm.interval(xub).midP();
    if ((nx - xm).sign(false) == 0 || (nx - xub).sign(false) == 0)
      break;
    int sm = value(nx).sign(false);
    if (sm == 0)
      xm = nx;
    else
      xub = nx;
  }
  Parameter nx = xlb.interval(xub);
  return nx.intervalWidth() < x.intervalWidth() ? nx : x;
}

PPoly2::PPoly2 (int nt, Parameter *a, int *m) : nt(nt)
{
  int k = nt*2;
  this->a = new Parameter [nt];
  this->m = new int [k];
  for (int i = 0; i < nt; ++i)
    this->a[i] = a[i];
  for (int i = 0; i < k; ++i)
    this->m[i] = m[i];
  setDegree();
}

PPoly2::PPoly2 (const PPoly2 &p) : nt(p.nt)
{
  if (p.a) {
    int k = nt*2;
    a = new Parameter [nt];
    m = new int [k];
    for (int i = 0; i < nt; ++i)
      a[i] = p.a[i];
    for (int i = 0; i < k; ++i)
      m[i] = p.m[i];
  }
  else {
    a = 0;
    m = 0;
  }
  d = p.d;
}

void PPoly2::setDegree ()
{
  d = 0;
  for (int i = 0; i < nt; ++i) {
    int di = 0;
    for (int j = 0; j < 2; ++j)
      di += m[2*i+j];
    if (d < di)
      d = di;
  }
}

PPoly2 & PPoly2::operator= (const PPoly2 &p)
{
  nt = p.nt;
  delete [] a;
  delete [] m;
  if (p.a) {
    int k = nt*2;
    a = new Parameter [nt];
    m = new int [k];
    for (int i = 0; i < nt; ++i)
      a[i] = p.a[i];
    for (int i = 0; i < k; ++i)
      m[i] = p.m[i];
  }
  else {
    a = 0;
    m = 0;
  }
  d = p.d;
  return *this;
}

Parameter PPoly2::value (Parameter *x) const
{
  // ??? vjm sentinal problem with y+z
  Parameter y;
  int *mp = m;
  for (int i = 0; i < nt; ++i) {
    Parameter z = a[i];
    for (int j = 0; j < 2; ++j, ++mp)
      for (int k = 0; k < *mp; ++k)
	z = z*x[j];
    if (i == 0)
      y = z;
    else
      y = y + z;
  }
  return y;
}

PPoly2 PPoly2::operator* (const PPoly2 &b) const 
{
  int cnt = 0;
  // Parameter ca[nt*b.nt];
  vector<Parameter> ca(nt*b.nt);
  int cm[2*nt*b.nt];
  int exps[2];
  for (int i = 0; i < nt; i++) {
    int *mp = m + i * 2;
    for (int j = 0; j < b.nt; j++) {
      Parameter coef = a[i] * b.a[j];
      int *bmp = b.m + j * 2;
      for (int k = 0; k < 2; k++)
        exps[k] = mp[k] + bmp[k];
      int l;
      for (l = 0; l < cnt; l++) {
        int *cmp = cm + l * 2;
        int h;
        for (h = 0; h < 2; h++)
          if (cmp[h] != exps[h])
            break;
        if (h == 2) {
          ca[l] = ca[l] + coef;
          break;
        }
      }
      if (l == cnt) {
        int *cmp = cm + l * 2;
        for (int h = 0; h < 2; h++)
          cmp[h] = exps[h];
        ca[l] = coef;
        cnt++;
      }
    }
  }
  // return PPoly2(cnt, ca, cm);
  return PPoly2(cnt, &ca[0], cm);
}

PPoly2 PPoly2::plusCtimes (double c, const PPoly2 &b) const {
  int pnt = nt;
  vector<Parameter> pa(nt + b.nt);
  int pm[2 * (nt + b.nt)];

  for (int i = 0; i < nt; i++)
    pa[i] = a[i];
  int ntnv = nt * 2;
  for (int i = 0; i < ntnv; i++)
    pm[i] = m[i];

  int *bmp = b.m;
  for (int i = 0; i < b.nt; i++) {
    int j;
    for (j = 0; j < nt; j++) {
      int *mp = m + j * 2;
      int k;
      for (k = 0; k < 2; k++)
        if (mp[k] != bmp[k])
          break;
      if (k == 2) {
        pa[j] = pa[j] + c * b.a[i];
        break;
      }
    }
    if (j == nt) {
      pa[pnt] = c * b.a[i];
      int *pmp = pm + pnt * 2;
      for (int k = 0; k < 2; k++)
        pmp[k] = bmp[k];
      pnt++;
    }
    bmp += 2;
  }

  return PPoly2(pnt, &pa[0], pm);
}
}

