// acp.h provides the acp::Parameter (numerical) class.
#include "acp.h"
// pv.h provides basic 2D and 3D vector of Parameter called acp::PV2 and
// acp::PV3 with the usual vector operations.
#include "pv.h"
// object.h provides acp::Object, acp::Predicate and macros to create
// predicates.
#include "object.h"
// Actually, object.h includes pv.h which includes acp.h, so it is
// sufficient to include object.h
using namespace acp;


// ABSTRACT GEOMETRIC OBJECTS

// Abstract Point class.
class Point : public Object<PV3> {
public:
  PV3 getP () { return get(); }
  virtual ~Point () {}
};

// Abstract plane class.
class Plane : public Object<PV3>, public Object<Parameter> {
public:
  PV3 getV () { return Object<PV3>::get(); }
  Parameter getD () { return Object<Parameter>::get(); }
  virtual ~Plane () {}
};
  

// PREDICATES

// Here is our first predicate class.
// Example invocation:  "if (PlaneSide(plane, point) > 0)".
// Sorry this is a clunky macro.
// int PlaneSide::side() must be defined in predicates3d.cc.
Predicate2(PlaneSide, Plane*, plane, Point*, point);

// Orientation of tetrahedron.
Predicate4(Orient3D, Point*, a, Point*, b, Point*, c, Point*, d);

// Insphere test.
Predicate5(InSphere, Point*, a, Point*, b, Point*, c, Point*, d, Point*, e);


// CONCRETE GEOMETRIC OBJECTS

// Point created from input coordinates.
class InputPoint : public Point {
  void calculate () {} // No calculation necessary.

public:
  InputPoint (double x, double y, double z) { set(PV3::input(x, y, z)); }
};

// Plane defined by triangle of points.
class TrianglePlane : public Plane {
  Point *a, *b, *c;
  void calculate ();
  
public:
  TrianglePlane (Point *a, Point *b, Point *c) : a(a), b(b), c(c) {
    // Object<Parameter>::safe_calculate();
  }
};

// Point defined as the projection of a point onto a plane
class PlanePoint : public Point {
  Plane *plane;
  Point *point;
  void calculate ();
  
public:
  PlanePoint (Plane *plane, Point *point) : plane(plane), point(point) {
    // safe_calculate();
  }
};

class IntersectionPoint : public Point {
  void calculate ();

protected:
  Point *tail;
  Point *head;
  Plane *plane;
  
public:
  IntersectionPoint (Point *tail, Point *head, Plane *plane)
    : tail(tail), head(head), plane(plane) {
    // safe_calculate();
  }
};

class IntersectionPosition : public IntersectionPoint, public Object<Parameter> {
  void calculate ();

public:
  IntersectionPosition (Point *tail, Point *head, Plane *plane)
    : IntersectionPoint(tail, head, plane) {
    // Object<Parameter>::safe_calculate();
  }
  Parameter getS () { return Object<Parameter>::get(); }
};

